@@ -1,10 +1,10 @@
 from os import path
 
-__version__ = '0.3.6'
+__version__ = '0.6.0'
 
 package_dir = path.abspath(path.dirname(__file__))
 
 
-from alembic import op
-from alembic import context
+from . import op
+from . import context
 
