@@ -1,13 +1,16 @@
 """Provide the 'autogenerate' feature which can produce migration operations
 automatically."""
 
-from alembic import util
-from sqlalchemy.engine.reflection import Inspector
-from sqlalchemy.util import OrderedSet
-from sqlalchemy import schema, types as sqltypes
+import logging
 import re
 
-import logging
+from sqlalchemy.engine.reflection import Inspector
+from sqlalchemy.util import OrderedSet
+from sqlalchemy import schema as sa_schema, types as sqltypes
+
+from . import util
+from .compat import string_types
+
 log = logging.getLogger(__name__)
 
 ###################################################
@@ -73,12 +76,15 @@ def compare_metadata(context, metadata):
             Table(u'bar', MetaData(bind=None),
                 Column(u'data', VARCHAR(), table=<bar>), schema=None)),
           ( 'add_column',
+            None,
             'foo',
             Column('data', Integer(), table=<foo>)),
           ( 'remove_column',
+            None,
             'foo',
             Column(u'old_data', VARCHAR(), table=None)),
           [ ( 'modify_nullable',
+              None,
               'foo',
               u'x',
               { 'existing_server_default': None,
@@ -102,10 +108,26 @@ def compare_metadata(context, metadata):
 # top level
 
 def _produce_migration_diffs(context, template_args,
-                                imports, include_symbol=None):
+                                imports, include_symbol=None,
+                                include_object=None,
+                                include_schemas=False):
     opts = context.opts
     metadata = opts['target_metadata']
+    include_object = opts.get('include_object', include_object)
     include_symbol = opts.get('include_symbol', include_symbol)
+    include_schemas = opts.get('include_schemas', include_schemas)
+
+    object_filters = []
+    if include_symbol:
+        def include_symbol_filter(object, name, type_, reflected, compare_to):
+            if type_ == "table":
+                return include_symbol(name, object.schema)
+            else:
+                return True
+        object_filters.append(include_symbol_filter)
+    if include_object:
+        object_filters.append(include_object)
+
 
     if metadata is None:
         raise util.CommandError(
@@ -118,7 +140,7 @@ def _produce_migration_diffs(context, template_args,
 
     diffs = []
     _produce_net_changes(connection, metadata, diffs,
-                                autogen_context, include_symbol)
+                                autogen_context, object_filters, include_schemas)
     template_args[opts['upgrade_token']] = \
             _indent(_produce_upgrade_commands(diffs, autogen_context))
     template_args[opts['downgrade_token']] = \
@@ -146,58 +168,85 @@ def _indent(text):
 ###################################################
 # walk structures
 
+
+def _run_filters(object_, name, type_, reflected, compare_to, object_filters):
+    for fn in object_filters:
+        if not fn(object_, name, type_, reflected, compare_to):
+            return False
+    else:
+        return True
+
 def _produce_net_changes(connection, metadata, diffs, autogen_context,
-                            include_symbol=None):
+                            object_filters=(),
+                            include_schemas=False):
     inspector = Inspector.from_engine(connection)
     # TODO: not hardcode alembic_version here ?
-    conn_table_names = set(inspector.get_table_names()).\
-                            difference(['alembic_version'])
+    conn_table_names = set()
+    if include_schemas:
+        schemas = set(inspector.get_schema_names())
+        # replace default schema name with None
+        schemas.discard("information_schema")
+        # replace the "default" schema with None
+        schemas.add(None)
+        schemas.discard(connection.dialect.default_schema_name)
+    else:
+        schemas = [None]
 
+    for s in schemas:
+        tables = set(inspector.get_table_names(schema=s)).\
+                difference(['alembic_version'])
+        conn_table_names.update(zip([s] * len(tables), tables))
 
-    metadata_table_names = OrderedSet([table.name
+    metadata_table_names = OrderedSet([(table.schema, table.name)
                                 for table in metadata.sorted_tables])
 
-    if include_symbol:
-        conn_table_names = set(name for name in conn_table_names
-                            if include_symbol(name))
-        metadata_table_names = OrderedSet(name for name in metadata_table_names
-                                if include_symbol(name))
-
     _compare_tables(conn_table_names, metadata_table_names,
+                    object_filters,
                     inspector, metadata, diffs, autogen_context)
 
 def _compare_tables(conn_table_names, metadata_table_names,
+                    object_filters,
                     inspector, metadata, diffs, autogen_context):
-    for tname in metadata_table_names.difference(conn_table_names):
-        diffs.append(("add_table", metadata.tables[tname]))
-        log.info("Detected added table %r", tname)
 
-    removal_metadata = schema.MetaData()
-    for tname in conn_table_names.difference(metadata_table_names):
-        exists = tname in removal_metadata.tables
-        t = schema.Table(tname, removal_metadata)
+    for s, tname in metadata_table_names.difference(conn_table_names):
+        name = '%s.%s' % (s, tname) if s else tname
+        metadata_table = metadata.tables[sa_schema._get_table_key(tname, s)]
+        if _run_filters(metadata_table, tname, "table", False, None, object_filters):
+            diffs.append(("add_table", metadata.tables[name]))
+            log.info("Detected added table %r", name)
+
+    removal_metadata = sa_schema.MetaData()
+    for s, tname in conn_table_names.difference(metadata_table_names):
+        name = sa_schema._get_table_key(tname, s)
+        exists = name in removal_metadata.tables
+        t = sa_schema.Table(tname, removal_metadata, schema=s)
         if not exists:
             inspector.reflecttable(t, None)
-        diffs.append(("remove_table", t))
-        log.info("Detected removed table %r", tname)
+        if _run_filters(t, tname, "table", True, None, object_filters):
+            diffs.append(("remove_table", t))
+            log.info("Detected removed table %r", name)
 
     existing_tables = conn_table_names.intersection(metadata_table_names)
 
-    conn_column_info = dict(
-        (tname,
-            dict(
-                (rec["name"], rec)
-                for rec in inspector.get_columns(tname)
-            )
-        )
-        for tname in existing_tables
-    )
+    existing_metadata = sa_schema.MetaData()
+    conn_column_info = {}
+    for s, tname in existing_tables:
+        name = sa_schema._get_table_key(tname, s)
+        exists = name in existing_metadata.tables
+        t = sa_schema.Table(tname, existing_metadata, schema=s)
+        if not exists:
+            inspector.reflecttable(t, None)
+        conn_column_info[(s, tname)] = t
 
-    for tname in sorted(existing_tables):
-        _compare_columns(tname,
-                conn_column_info[tname],
-                metadata.tables[tname],
-                diffs, autogen_context)
+    for s, tname in sorted(existing_tables):
+        name = '%s.%s' % (s, tname) if s else tname
+        metadata_table = metadata.tables[name]
+        conn_table = existing_metadata.tables[name]
+        if _run_filters(metadata_table, tname, "table", False, conn_table, object_filters):
+            _compare_columns(s, tname, object_filters,
+                    conn_table,
+                    metadata_table,
+                    diffs, autogen_context)
 
     # TODO:
     # index add/drop
@@ -207,44 +256,54 @@ def _compare_tables(conn_table_names, metadata_table_names,
 ###################################################
 # element comparison
 
-def _compare_columns(tname, conn_table, metadata_table,
+def _compare_columns(schema, tname, object_filters, conn_table, metadata_table,
                                 diffs, autogen_context):
+    name = '%s.%s' % (schema, tname) if schema else tname
     metadata_cols_by_name = dict((c.name, c) for c in metadata_table.c)
-    conn_col_names = set(conn_table)
-    metadata_col_names = set(metadata_cols_by_name)
+    conn_col_names = dict((c.name, c) for c in conn_table.c)
+    metadata_col_names = OrderedSet(sorted(metadata_cols_by_name))
+
 
     for cname in metadata_col_names.difference(conn_col_names):
-        diffs.append(
-            ("add_column", tname, metadata_cols_by_name[cname])
-        )
-        log.info("Detected added column '%s.%s'", tname, cname)
+        if _run_filters(metadata_cols_by_name[cname], cname,
+                                "column", False, None, object_filters):
+            diffs.append(
+                ("add_column", schema, tname, metadata_cols_by_name[cname])
+            )
+            log.info("Detected added column '%s.%s'", name, cname)
 
-    for cname in conn_col_names.difference(metadata_col_names):
-        diffs.append(
-            ("remove_column", tname, schema.Column(
-                cname,
-                conn_table[cname]['type'],
-                nullable=conn_table[cname]['nullable'],
-                server_default=conn_table[cname]['default']
-            ))
-        )
-        log.info("Detected removed column '%s.%s'", tname, cname)
+    for cname in set(conn_col_names).difference(metadata_col_names):
+        rem_col = sa_schema.Column(
+                    cname,
+                    conn_table.c[cname].type,
+                    nullable=conn_table.c[cname].nullable,
+                    server_default=conn_table.c[cname].server_default
+                )
+        if _run_filters(rem_col, cname,
+                                "column", True, None, object_filters):
+            diffs.append(
+                ("remove_column", schema, tname, rem_col)
+            )
+            log.info("Detected removed column '%s.%s'", name, cname)
 
     for colname in metadata_col_names.intersection(conn_col_names):
-        metadata_col = metadata_table.c[colname]
-        conn_col = conn_table[colname]
+        metadata_col = metadata_cols_by_name[colname]
+        conn_col = conn_table.c[colname]
+        if not _run_filters(
+                    metadata_col, colname, "column", False, conn_col, object_filters):
+            continue
         col_diff = []
-        _compare_type(tname, colname,
+        _compare_type(schema, tname, colname,
             conn_col,
             metadata_col,
             col_diff, autogen_context
         )
-        _compare_nullable(tname, colname,
+        _compare_nullable(schema, tname, colname,
             conn_col,
             metadata_col.nullable,
             col_diff, autogen_context
         )
-        _compare_server_default(tname, colname,
+        _compare_server_default(schema, tname, colname,
             conn_col,
             metadata_col,
             col_diff, autogen_context
@@ -252,16 +311,16 @@ def _compare_columns(tname, conn_table, metadata_table,
         if col_diff:
             diffs.append(col_diff)
 
-def _compare_nullable(tname, cname, conn_col,
+def _compare_nullable(schema, tname, cname, conn_col,
                             metadata_col_nullable, diffs,
                             autogen_context):
-    conn_col_nullable = conn_col['nullable']
+    conn_col_nullable = conn_col.nullable
     if conn_col_nullable is not metadata_col_nullable:
         diffs.append(
-            ("modify_nullable", tname, cname,
+            ("modify_nullable", schema, tname, cname,
                 {
-                    "existing_type": conn_col['type'],
-                    "existing_server_default": conn_col['default'],
+                    "existing_type": conn_col.type,
+                    "existing_server_default": conn_col.server_default,
                 },
                 conn_col_nullable,
                 metadata_col_nullable),
@@ -272,11 +331,11 @@ def _compare_nullable(tname, cname, conn_col,
             cname
         )
 
-def _compare_type(tname, cname, conn_col,
+def _compare_type(schema, tname, cname, conn_col,
                             metadata_col, diffs,
                             autogen_context):
 
-    conn_type = conn_col['type']
+    conn_type = conn_col.type
     metadata_type = metadata_col.type
     if conn_type._type_affinity is sqltypes.NullType:
         log.info("Couldn't determine database type "
@@ -292,10 +351,10 @@ def _compare_type(tname, cname, conn_col,
     if isdiff:
 
         diffs.append(
-            ("modify_type", tname, cname,
+            ("modify_type", schema, tname, cname,
                     {
-                        "existing_nullable": conn_col['nullable'],
-                        "existing_server_default": conn_col['default'],
+                        "existing_nullable": conn_col.nullable,
+                        "existing_server_default": conn_col.server_default,
                     },
                     conn_type,
                     metadata_type),
@@ -304,26 +363,29 @@ def _compare_type(tname, cname, conn_col,
             conn_type, metadata_type, tname, cname
         )
 
-def _compare_server_default(tname, cname, conn_col, metadata_col,
+def _compare_server_default(schema, tname, cname, conn_col, metadata_col,
                                 diffs, autogen_context):
 
     metadata_default = metadata_col.server_default
-    conn_col_default = conn_col['default']
+    conn_col_default = conn_col.server_default
     if conn_col_default is None and metadata_default is None:
         return False
     rendered_metadata_default = _render_server_default(
                             metadata_default, autogen_context)
+    rendered_conn_default = conn_col.server_default.arg.text \
+                            if conn_col.server_default else None
     isdiff = autogen_context['context']._compare_server_default(
                         conn_col, metadata_col,
-                        rendered_metadata_default
+                        rendered_metadata_default,
+                        rendered_conn_default
                     )
     if isdiff:
-        conn_col_default = conn_col['default']
+        conn_col_default = rendered_conn_default
         diffs.append(
-            ("modify_default", tname, cname,
+            ("modify_default", schema, tname, cname,
                 {
-                    "existing_nullable": conn_col['nullable'],
-                    "existing_type": conn_col['type'],
+                    "existing_nullable": conn_col.nullable,
+                    "existing_type": conn_col.type,
                 },
                 conn_col_default,
                 metadata_default),
@@ -382,7 +444,7 @@ def _invoke_adddrop_command(updown, args, autogen_context):
         return cmd_callables[0](*cmd_args)
 
 def _invoke_modify_command(updown, args, autogen_context):
-    tname, cname = args[0][1:3]
+    sname, tname, cname = args[0][1:4]
     kw = {}
 
     _arg_struct = {
@@ -391,7 +453,7 @@ def _invoke_modify_command(updown, args, autogen_context):
         "modify_default": ("existing_server_default", "server_default"),
     }
     for diff in args:
-        diff_kw = diff[3]
+        diff_kw = diff[4]
         for arg in ("existing_type", \
                 "existing_nullable", \
                 "existing_server_default"):
@@ -409,44 +471,64 @@ def _invoke_modify_command(updown, args, autogen_context):
         kw.pop("existing_nullable", None)
     if "server_default" in kw:
         kw.pop("existing_server_default", None)
-    return _modify_col(tname, cname, autogen_context, **kw)
+    return _modify_col(tname, cname, autogen_context, schema=sname, **kw)
 
 ###################################################
 # render python
 
 def _add_table(table, autogen_context):
-    return "%(prefix)screate_table(%(tablename)r,\n%(args)s\n)" % {
+    text = "%(prefix)screate_table(%(tablename)r,\n%(args)s" % {
         'tablename': table.name,
         'prefix': _alembic_autogenerate_prefix(autogen_context),
         'args': ',\n'.join(
-            [_render_column(col, autogen_context) for col in table.c] +
+            [col for col in
+                [_render_column(col, autogen_context) for col in table.c]
+            if col] +
             sorted([rcons for rcons in
                 [_render_constraint(cons, autogen_context) for cons in
                     table.constraints]
                 if rcons is not None
             ])
-        ),
+        )
     }
+    if table.schema:
+        text += ",\nschema=%r" % table.schema
+    for k in sorted(table.kwargs):
+        text += ",\n%s=%r" % (k.replace(" ", "_"), table.kwargs[k])
+    text += "\n)"
+    return text
 
 def _drop_table(table, autogen_context):
-    return "%(prefix)sdrop_table(%(tname)r)" % {
+    text = "%(prefix)sdrop_table(%(tname)r" % {
             "prefix": _alembic_autogenerate_prefix(autogen_context),
             "tname": table.name
         }
+    if table.schema:
+        text += ", schema=%r" % table.schema
+    text += ")"
+    return text
 
-def _add_column(tname, column, autogen_context):
-    return "%(prefix)sadd_column(%(tname)r, %(column)s)" % {
+def _add_column(schema, tname, column, autogen_context):
+    text = "%(prefix)sadd_column(%(tname)r, %(column)s" % {
             "prefix": _alembic_autogenerate_prefix(autogen_context),
             "tname": tname,
             "column": _render_column(column, autogen_context)
             }
+    if schema:
+        text += ", schema=%r" % schema
+    text += ")"
+    return text
 
-def _drop_column(tname, column, autogen_context):
-    return "%(prefix)sdrop_column(%(tname)r, %(cname)r)" % {
+def _drop_column(schema, tname, column, autogen_context):
+    text = "%(prefix)sdrop_column(%(tname)r, %(cname)r" % {
             "prefix": _alembic_autogenerate_prefix(autogen_context),
             "tname": tname,
             "cname": column.name
             }
+    if schema:
+        text += ", schema=%r" % schema
+    text += ")"
+    return text
 
 def _modify_col(tname, cname,
                 autogen_context,
@@ -455,7 +537,8 @@ def _modify_col(tname, cname,
                 nullable=None,
                 existing_type=None,
                 existing_nullable=None,
-                existing_server_default=False):
+                existing_server_default=False,
+                schema=None):
     sqla_prefix = _sqlalchemy_autogenerate_prefix(autogen_context)
     indent = " " * 11
     text = "%(prefix)salter_column(%(tname)r, %(cname)r" % {
@@ -466,9 +549,10 @@ def _modify_col(tname, cname,
     text += ",\n%sexisting_type=%s" % (indent,
                     _repr_type(sqla_prefix, existing_type, autogen_context))
     if server_default is not False:
-        text += ",\n%sserver_default=%s" % (indent,
-                        _render_server_default(
-                                server_default, autogen_context),)
+        rendered = _render_server_default(
+                                server_default, autogen_context)
+        text += ",\n%sserver_default=%s" % (indent, rendered)
+
     if type_ is not None:
         text += ",\n%stype_=%s" % (indent,
                         _repr_type(sqla_prefix, type_, autogen_context))
@@ -479,12 +563,13 @@ def _modify_col(tname, cname,
         text += ",\n%sexisting_nullable=%r" % (
                         indent, existing_nullable)
     if existing_server_default:
-        text += ",\n%sexisting_server_default=%s" % (
-                        indent,
-                        _render_server_default(
+        rendered = _render_server_default(
                             existing_server_default,
-                            autogen_context),
-                    )
+                            autogen_context)
+        text += ",\n%sexisting_server_default=%s" % (
+                        indent, rendered)
+    if schema:
+        text += ",\n%sschema=%r" % (indent, schema)
     text += ")"
     return text
 
@@ -494,13 +579,33 @@ def _sqlalchemy_autogenerate_prefix(autogen_context):
 def _alembic_autogenerate_prefix(autogen_context):
     return autogen_context['opts']['alembic_module_prefix'] or ''
 
+
+def _user_defined_render(type_, object_, autogen_context):
+    if 'opts' in autogen_context and \
+            'render_item' in autogen_context['opts']:
+        render = autogen_context['opts']['render_item']
+        if render:
+            rendered = render(type_, object_, autogen_context)
+            if rendered is not False:
+                return rendered
+    return False
+
 def _render_column(column, autogen_context):
+    rendered = _user_defined_render("column", column, autogen_context)
+    if rendered is not False:
+        return rendered
+
     opts = []
     if column.server_default:
-        opts.append(("server_default",
-                    _render_server_default(
+        rendered = _render_server_default(
                             column.server_default, autogen_context
-                    )))
+                    )
+        if rendered:
+            opts.append(("server_default", rendered))
+
+    if not column.autoincrement:
+        opts.append(("autoincrement", column.autoincrement))
+
     if column.nullable is not None:
         opts.append(("nullable", column.nullable))
 
@@ -514,13 +619,17 @@ def _render_column(column, autogen_context):
     }
 
 def _render_server_default(default, autogen_context):
-    if isinstance(default, schema.DefaultClause):
-        if isinstance(default.arg, basestring):
+    rendered = _user_defined_render("server_default", default, autogen_context)
+    if rendered is not False:
+        return rendered
+
+    if isinstance(default, sa_schema.DefaultClause):
+        if isinstance(default.arg, string_types):
             default = default.arg
         else:
             default = str(default.arg.compile(
                             dialect=autogen_context['dialect']))
-    if isinstance(default, basestring):
+    if isinstance(default, string_types):
         # TODO: this is just a hack to get
         # tests to pass until we figure out
         # WTF sqlite is doing
@@ -530,6 +639,10 @@ def _render_server_default(default, autogen_context):
         return None
 
 def _repr_type(prefix, type_, autogen_context):
+    rendered = _user_defined_render("type", type_, autogen_context)
+    if rendered is not False:
+        return rendered
+
     mod = type(type_).__module__
     imports = autogen_context.get('imports', None)
     if mod.startswith("sqlalchemy.dialects"):
@@ -548,6 +661,10 @@ def _render_constraint(constraint, autogen_context):
         return None
 
 def _render_primary_key(constraint, autogen_context):
+    rendered = _user_defined_render("primary_key", constraint, autogen_context)
+    if rendered is not False:
+        return rendered
+
     opts = []
     if constraint.name:
         opts.append(("name", repr(constraint.name)))
@@ -559,16 +676,49 @@ def _render_primary_key(constraint, autogen_context):
         ),
     }
 
+def _fk_colspec(fk, metadata_schema):
+    """Implement a 'safe' version of ForeignKey._get_colspec() that
+    never tries to resolve the remote table.
+
+    """
+    if metadata_schema is None:
+        return fk._get_colspec()
+    else:
+        # need to render schema breaking up tokens by hand, since the
+        # ForeignKeyConstraint here may not actually have a remote
+        # Table present
+        tokens = fk._colspec.split(".")
+        # no schema in the colspec, render it
+        if len(tokens) == 2:
+            return "%s.%s" % (metadata_schema, fk._colspec)
+        else:
+            return fk._colspec
+
 def _render_foreign_key(constraint, autogen_context):
+    rendered = _user_defined_render("foreign_key", constraint, autogen_context)
+    if rendered is not False:
+        return rendered
+
     opts = []
     if constraint.name:
         opts.append(("name", repr(constraint.name)))
-    # TODO: deferrable, initially, etc.
+    if constraint.onupdate:
+        opts.append(("onupdate", repr(constraint.onupdate)))
+    if constraint.ondelete:
+        opts.append(("ondelete", repr(constraint.ondelete)))
+    if constraint.initially:
+        opts.append(("initially", repr(constraint.initially)))
+    if constraint.deferrable:
+        opts.append(("deferrable", repr(constraint.deferrable)))
+    if constraint.use_alter:
+        opts.append(("use_alter", repr(constraint.use_alter)))
+
+    apply_metadata_schema = constraint.parent.metadata.schema
     return "%(prefix)sForeignKeyConstraint([%(cols)s], "\
             "[%(refcols)s], %(args)s)" % {
         "prefix": _sqlalchemy_autogenerate_prefix(autogen_context),
         "cols": ", ".join("'%s'" % f.parent.key for f in constraint.elements),
-        "refcols": ", ".join(repr(f._get_colspec())
+        "refcols": ", ".join(repr(_fk_colspec(f, apply_metadata_schema))
                             for f in constraint.elements),
         "args": ", ".join(
             ["%s=%s" % (kwname, val) for kwname, val in opts]
@@ -576,6 +726,10 @@ def _render_foreign_key(constraint, autogen_context):
     }
 
 def _render_check_constraint(constraint, autogen_context):
+    rendered = _user_defined_render("check", constraint, autogen_context)
+    if rendered is not False:
+        return rendered
+
     # detect the constraint being part of
     # a parent type which is probably in the Table already.
     # ideally SQLAlchemy would give us more of a first class
@@ -588,8 +742,10 @@ def _render_check_constraint(constraint, autogen_context):
     opts = []
     if constraint.name:
         opts.append(("name", repr(constraint.name)))
-    return "%(prefix)sCheckConstraint(%(sqltext)r)" % {
+    return "%(prefix)sCheckConstraint(%(sqltext)r%(opts)s)" % {
             "prefix": _sqlalchemy_autogenerate_prefix(autogen_context),
+            "opts": ", " + (", ".join("%s=%s" % (k, v)
+                            for k, v in opts)) if opts else "",
             "sqltext": str(
                 constraint.sqltext.compile(
                     dialect=autogen_context['dialect']
@@ -598,6 +754,10 @@ def _render_check_constraint(constraint, autogen_context):
         }
 
 def _render_unique_constraint(constraint, autogen_context):
+    rendered = _user_defined_render("unique", constraint, autogen_context)
+    if rendered is not False:
+        return rendered
+
     opts = []
     if constraint.name:
         opts.append(("name", "'%s'" % constraint.name))
@@ -608,8 +768,8 @@ def _render_unique_constraint(constraint, autogen_context):
         "prefix": _sqlalchemy_autogenerate_prefix(autogen_context)
         }
 _constraint_renderers = {
-    schema.PrimaryKeyConstraint: _render_primary_key,
-    schema.ForeignKeyConstraint: _render_foreign_key,
-    schema.UniqueConstraint: _render_unique_constraint,
-    schema.CheckConstraint: _render_check_constraint
+    sa_schema.PrimaryKeyConstraint: _render_primary_key,
+    sa_schema.ForeignKeyConstraint: _render_foreign_key,
+    sa_schema.UniqueConstraint: _render_unique_constraint,
+    sa_schema.CheckConstraint: _render_check_constraint
 }
