@@ -1,9 +1,11 @@
-from alembic import util
-from alembic.ddl import impl
-from sqlalchemy.types import NULLTYPE, Integer
-from sqlalchemy import schema, sql
 from contextlib import contextmanager
-import alembic
+
+from sqlalchemy.types import NULLTYPE, Integer
+from sqlalchemy import schema as sa_schema
+
+from . import util
+from .compat import string_types
+from .ddl import impl
 
 __all__ = ('Operations',)
 
@@ -45,27 +47,44 @@ class Operations(object):
     @classmethod
     @contextmanager
     def context(cls, migration_context):
+        from .op import _install_proxy, _remove_proxy
         op = Operations(migration_context)
-        alembic.op._install_proxy(op)
+        _install_proxy(op)
         yield op
-        alembic.op._remove_proxy()
+        _remove_proxy()
+
+
+    def _primary_key_constraint(self, name, table_name, cols, schema=None):
+        m = sa_schema.MetaData()
+        columns = [sa_schema.Column(n, NULLTYPE) for n in cols]
+        t1 = sa_schema.Table(table_name, m,
+                *columns,
+                schema=schema)
+        p = sa_schema.PrimaryKeyConstraint(*columns, name=name)
+        t1.append_constraint(p)
+        return p
 
     def _foreign_key_constraint(self, name, source, referent,
                                     local_cols, remote_cols,
-                                    onupdate=None, ondelete=None):
-        m = schema.MetaData()
+                                    onupdate=None, ondelete=None,
+                                    source_schema=None, referent_schema=None):
+        m = sa_schema.MetaData()
         if source == referent:
             t1_cols = local_cols + remote_cols
         else:
             t1_cols = local_cols
-            schema.Table(referent, m,
-                *[schema.Column(n, NULLTYPE) for n in remote_cols])
+            sa_schema.Table(referent, m,
+                    *[sa_schema.Column(n, NULLTYPE) for n in remote_cols],
+                    schema=referent_schema)
 
-        t1 = schema.Table(source, m,
-                *[schema.Column(n, NULLTYPE) for n in t1_cols])
+        t1 = sa_schema.Table(source, m,
+                *[sa_schema.Column(n, NULLTYPE) for n in t1_cols],
+                schema=source_schema)
 
-        f = schema.ForeignKeyConstraint(local_cols,
-                                            ["%s.%s" % (referent, n)
+        tname = "%s.%s" % (referent_schema, referent) if referent_schema \
+                else referent
+        f = sa_schema.ForeignKeyConstraint(local_cols,
+                                            ["%s.%s" % (tname, n)
                                             for n in remote_cols],
                                             name=name,
                                             onupdate=onupdate,
@@ -75,59 +94,65 @@ class Operations(object):
 
         return f
 
-    def _unique_constraint(self, name, source, local_cols, **kw):
-        t = schema.Table(source, schema.MetaData(),
-                    *[schema.Column(n, NULLTYPE) for n in local_cols])
+    def _unique_constraint(self, name, source, local_cols, schema=None, **kw):
+        t = sa_schema.Table(source, sa_schema.MetaData(),
+                    *[sa_schema.Column(n, NULLTYPE) for n in local_cols],
+                    schema=schema)
         kw['name'] = name
-        uq = schema.UniqueConstraint(*[t.c[n] for n in local_cols], **kw)
+        uq = sa_schema.UniqueConstraint(*[t.c[n] for n in local_cols], **kw)
         # TODO: need event tests to ensure the event
         # is fired off here
         t.append_constraint(uq)
         return uq
 
-    def _check_constraint(self, name, source, condition, **kw):
-        t = schema.Table(source, schema.MetaData(),
-                    schema.Column('x', Integer))
-        ck = schema.CheckConstraint(condition, name=name, **kw)
+    def _check_constraint(self, name, source, condition, schema=None, **kw):
+        t = sa_schema.Table(source, sa_schema.MetaData(),
+                    sa_schema.Column('x', Integer), schema=schema)
+        ck = sa_schema.CheckConstraint(condition, name=name, **kw)
         t.append_constraint(ck)
         return ck
 
     def _table(self, name, *columns, **kw):
-        m = schema.MetaData()
-        t = schema.Table(name, m, *columns, **kw)
+        m = sa_schema.MetaData()
+        t = sa_schema.Table(name, m, *columns, **kw)
         for f in t.foreign_keys:
             self._ensure_table_for_fk(m, f)
         return t
 
     def _column(self, name, type_, **kw):
-        return schema.Column(name, type_, **kw)
+        return sa_schema.Column(name, type_, **kw)
 
-    def _index(self, name, tablename, columns, **kw):
-        t = schema.Table(tablename or 'no_table', schema.MetaData(),
-            *[schema.Column(n, NULLTYPE) for n in columns]
+    def _index(self, name, tablename, columns, schema=None, **kw):
+        t = sa_schema.Table(tablename or 'no_table', sa_schema.MetaData(),
+            *[sa_schema.Column(n, NULLTYPE) for n in columns],
+            schema=schema
         )
-        return schema.Index(name, *list(t.c), **kw)
+        return sa_schema.Index(name, *[t.c[n] for n in columns], **kw)
+
+    def _parse_table_key(self, table_key):
+        if '.' in table_key:
+            tokens = table_key.split('.')
+            sname = ".".join(tokens[0:-1])
+            tname = tokens[-1]
+        else:
+            tname = table_key
+            sname = None
+        return (sname, tname)
 
     def _ensure_table_for_fk(self, metadata, fk):
         """create a placeholder Table object for the referent of a
         ForeignKey.
 
         """
-        if isinstance(fk._colspec, basestring):
+        if isinstance(fk._colspec, string_types):
             table_key, cname = fk._colspec.rsplit('.', 1)
-            if '.' in table_key:
-                tokens = table_key.split('.')
-                sname = ".".join(tokens[0:-1])
-                tname = tokens[-1]
-            else:
-                tname = table_key
-                sname = None
+            sname, tname = self._parse_table_key(table_key)
             if table_key not in metadata.tables:
-                rel_t = schema.Table(tname, metadata, schema=sname)
+                rel_t = sa_schema.Table(tname, metadata, schema=sname)
             else:
                 rel_t = metadata.tables[table_key]
             if cname not in rel_t.c:
-                rel_t.append_column(schema.Column(cname, NULLTYPE))
+                rel_t.append_column(sa_schema.Column(cname, NULLTYPE))
 
     def get_context(self):
         """Return the :class:`.MigrationContext` object that's
@@ -142,7 +167,7 @@ class Operations(object):
 
         :param old_table_name: old name.
         :param new_table_name: new name.
-        :param schema: Optional, name of schema to operate within.
+        :param schema: Optional schema name to operate within.
 
         """
         self.impl.rename_table(
@@ -151,14 +176,18 @@ class Operations(object):
             schema=schema
         )
 
+    @util._with_legacy_names([('name', 'new_column_name')])
     def alter_column(self, table_name, column_name,
                         nullable=None,
                         server_default=False,
-                        name=None,
+                        new_column_name=None,
                         type_=None,
+                        autoincrement=None,
                         existing_type=None,
                         existing_server_default=False,
                         existing_nullable=None,
+                        existing_autoincrement=None,
+                        schema=None
     ):
         """Issue an "alter column" instruction using the
         current migration context.
@@ -196,14 +225,21 @@ class Operations(object):
          or :class:`~sqlalchemy.schema.DefaultClause` to indicate
          an alteration to the column's default value.
          Set to ``None`` to have the default removed.
-        :param name: Optional; specify a string name here to
+        :param new_column_name: Optional; specify a string name here to
          indicate the new name within a column rename operation.
-        :param type_: Optional; a :class:`~sqlalchemy.types.TypeEngine`
+
+         .. versionchanged:: 0.5.0
+            The ``name`` parameter is now named ``new_column_name``.
+            The old name will continue to function for backwards
+            compatibility.
+
+        :param ``type_``: Optional; a :class:`~sqlalchemy.types.TypeEngine`
          type object to specify a change to the column's type.
          For SQLAlchemy types that also indicate a constraint (i.e.
-         :class:`~sqlalchemy.types.Boolean`,
-         :class:`~sqlalchemy.types.Enum`),
+         :class:`~sqlalchemy.types.Boolean`, :class:`~sqlalchemy.types.Enum`),
          the constraint is also generated.
+        :param autoincrement: set the ``AUTO_INCREMENT`` flag of the column;
+         currently understood by the MySQL dialect.
         :param existing_type: Optional; a
          :class:`~sqlalchemy.types.TypeEngine`
          type object to specify the previous type.   This
@@ -211,8 +247,7 @@ class Operations(object):
          don't otherwise specify a new type, as well as for
          when nullability is being changed on a SQL Server
          column.  It is also used if the type is a so-called
-         SQLlchemy "schema" type which
-         may define a constraint (i.e.
+         SQLlchemy "schema" type which may define a constraint (i.e.
          :class:`~sqlalchemy.types.Boolean`,
          :class:`~sqlalchemy.types.Enum`),
          so that the constraint can be dropped.
@@ -223,6 +258,13 @@ class Operations(object):
         :param existing_nullable: Optional; the existing nullability
          of the column.  Required on MySQL if the existing nullability
          is not being changed; else MySQL sets this to NULL.
+        :param existing_autoincrement: Optional; the existing autoincrement
+         of the column.  Used for MySQL's system of altering a column
+         that specifies ``AUTO_INCREMENT``.
+        :param schema: Optional schema name to operate within.
+
+         .. versionadded:: 0.4.0
+
         """
 
         compiler = self.impl.dialect.statement_compiler(
@@ -230,13 +272,14 @@ class Operations(object):
                             None
                         )
         def _count_constraint(constraint):
-            return not isinstance(constraint, schema.PrimaryKeyConstraint) and \
+            return not isinstance(constraint, sa_schema.PrimaryKeyConstraint) and \
                 (not constraint._create_rule or
                     constraint._create_rule(compiler))
 
         if existing_type and type_:
             t = self._table(table_name,
-                        schema.Column(column_name, existing_type)
+                        sa_schema.Column(column_name, existing_type),
+                        schema=schema
                     )
             for constraint in t.constraints:
                 if _count_constraint(constraint):
@@ -245,20 +288,26 @@ class Operations(object):
         self.impl.alter_column(table_name, column_name,
             nullable=nullable,
             server_default=server_default,
-            name=name,
+            name=new_column_name,
             type_=type_,
+            schema=schema,
+            autoincrement=autoincrement,
             existing_type=existing_type,
             existing_server_default=existing_server_default,
             existing_nullable=existing_nullable,
+            existing_autoincrement=existing_autoincrement
         )
 
         if type_:
-            t = self._table(table_name, schema.Column(column_name, type_))
+            t = self._table(table_name,
+                        sa_schema.Column(column_name, type_),
+                        schema=schema
+                    )
             for constraint in t.constraints:
                 if _count_constraint(constraint):
                     self.impl.add_constraint(constraint)
 
-    def add_column(self, table_name, column):
+    def add_column(self, table_name, column, schema=None):
         """Issue an "add column" instruction using the current
         migration context.
 
@@ -301,16 +350,20 @@ class Operations(object):
         :param table_name: String name of the parent table.
         :param column: a :class:`sqlalchemy.schema.Column` object
          representing the new column.
+        :param schema: Optional schema name to operate within.
+
+         .. versionadded:: 0.4.0
 
         """
 
-        t = self._table(table_name, column)
+        t = self._table(table_name, column, schema=schema)
         self.impl.add_column(
             table_name,
-            column
+            column,
+            schema=schema
         )
         for constraint in t.constraints:
-            if not isinstance(constraint, schema.PrimaryKeyConstraint):
+            if not isinstance(constraint, sa_schema.PrimaryKeyConstraint):
                 self.impl.add_constraint(constraint)
 
     def drop_column(self, table_name, column_name, **kw):
@@ -323,6 +376,10 @@ class Operations(object):
 
         :param table_name: name of table
         :param column_name: name of column
+        :param schema: Optional schema name to operate within.
+
+         .. versionadded:: 0.4.0
+
         :param mssql_drop_check: Optional boolean.  When ``True``, on
          Microsoft SQL Server only, first
          drop the CHECK constraint on the column using a
@@ -345,8 +402,49 @@ class Operations(object):
         )
 
 
+    def create_primary_key(self, name, table_name, cols, schema=None):
+        """Issue a "create primary key" instruction using the current
+        migration context.
+
+        e.g.::
+
+            from alembic import op
+            op.create_primary_key(
+                        "pk_my_table", "my_table",
+                        ["id", "version"]
+                    )
+
+        This internally generates a :class:`~sqlalchemy.schema.Table` object
+        containing the necessary columns, then generates a new
+        :class:`~sqlalchemy.schema.PrimaryKeyConstraint`
+        object which it then associates with the :class:`~sqlalchemy.schema.Table`.
+        Any event listeners associated with this action will be fired
+        off normally.   The :class:`~sqlalchemy.schema.AddConstraint`
+        construct is ultimately used to generate the ALTER statement.
+
+        .. versionadded:: 0.5.0
+
+        :param name: Name of the primary key constraint.  The name is necessary
+         so that an ALTER statement can be emitted.  For setups that
+         use an automated naming scheme such as that described at
+         `NamingConventions <http://www.sqlalchemy.org/trac/wiki/UsageRecipes/NamingConventions>`_,
+         ``name`` here can be ``None``, as the event listener will
+         apply the name to the constraint object when it is associated
+         with the table.
+        :param table_name: String name of the target table.
+        :param cols: a list of string column names to be applied to the
+         primary key constraint.
+        :param schema: Optional schema name of the table.
+
+        """
+        self.impl.add_constraint(
+                    self._primary_key_constraint(name, table_name, cols,
+                                schema)
+                )
+
     def create_foreign_key(self, name, source, referent, local_cols,
-                           remote_cols, onupdate=None, ondelete=None):
+                           remote_cols, onupdate=None, ondelete=None,
+                           source_schema=None, referent_schema=None):
         """Issue a "create foreign key" instruction using the
         current migration context.
 
@@ -372,10 +470,8 @@ class Operations(object):
          ``name`` here can be ``None``, as the event listener will
          apply the name to the constraint object when it is associated
          with the table.
-        :param source: String name of the source table.  Currently
-         there is no support for dotted schema names.
-        :param referent: String name of the destination table. Currently
-         there is no support for dotted schema names.
+        :param source: String name of the source table.
+        :param referent: String name of the destination table.
         :param local_cols: a list of string column names in the
          source table.
         :param remote_cols: a list of string column names in the
@@ -386,16 +482,21 @@ class Operations(object):
         :param ondelete: Optional string. If set, emit ON DELETE <value> when
          issuing DDL for this constraint. Typical values include CASCADE,
          DELETE and RESTRICT.
+        :param source_schema: Optional schema name of the source table.
+        :param referent_schema: Optional schema name of the destination table.
 
         """
 
         self.impl.add_constraint(
                     self._foreign_key_constraint(name, source, referent,
                             local_cols, remote_cols,
-                            onupdate=onupdate, ondelete=ondelete)
+                            onupdate=onupdate, ondelete=ondelete,
+                            source_schema=source_schema,
+                            referent_schema=referent_schema)
                 )
 
-    def create_unique_constraint(self, name, source, local_cols, **kw):
+    def create_unique_constraint(self, name, source, local_cols,
+                                 schema=None, **kw):
         """Issue a "create unique constraint" instruction using the
         current migration context.
 
@@ -419,23 +520,27 @@ class Operations(object):
          ``name`` here can be ``None``, as the event listener will
          apply the name to the constraint object when it is associated
          with the table.
-        :param source: String name of the source table.  Currently
-         there is no support for dotted schema names.
+        :param source: String name of the source table. Dotted schema names are
+         supported.
         :param local_cols: a list of string column names in the
          source table.
         :param deferrable: optional bool. If set, emit DEFERRABLE or NOT DEFERRABLE when
          issuing DDL for this constraint.
         :param initially: optional string. If set, emit INITIALLY <value> when issuing DDL
          for this constraint.
+        :param schema: Optional schema name to operate within.
+
+         .. versionadded:: 0.4.0
 
         """
 
         self.impl.add_constraint(
                     self._unique_constraint(name, source, local_cols,
-                        **kw)
+                        schema=schema, **kw)
                 )
 
-    def create_check_constraint(self, name, source, condition, **kw):
+    def create_check_constraint(self, name, source, condition,
+                                schema=None, **kw):
         """Issue a "create check constraint" instruction using the
         current migration context.
 
@@ -462,18 +567,20 @@ class Operations(object):
          ``name`` here can be ``None``, as the event listener will
          apply the name to the constraint object when it is associated
          with the table.
-        :param source: String name of the source table.  Currently
-         there is no support for dotted schema names.
+        :param source: String name of the source table.
         :param condition: SQL expression that's the condition of the constraint.
          Can be a string or SQLAlchemy expression language structure.
         :param deferrable: optional bool. If set, emit DEFERRABLE or NOT DEFERRABLE when
          issuing DDL for this constraint.
         :param initially: optional string. If set, emit INITIALLY <value> when issuing DDL
          for this constraint.
+        :param schema: Optional schema name to operate within.
+
+         ..versionadded:: 0.4.0
 
         """
         self.impl.add_constraint(
-            self._check_constraint(name, source, condition, **kw)
+            self._check_constraint(name, source, condition, schema=schema, **kw)
         )
 
     def create_table(self, name, *columns, **kw):
@@ -515,19 +622,16 @@ class Operations(object):
          the table, as well as optional :class:`~sqlalchemy.schema.Constraint`
          objects
          and :class:`~.sqlalchemy.schema.Index` objects.
-        :param emit_events: if ``True``, emit ``before_create`` and
-         ``after_create`` events when the table is being created.  In
-         particular, the Postgresql ENUM type will emit a CREATE TYPE within
-         these events.
+        :param schema: Optional schema name to operate within.
         :param \**kw: Other keyword arguments are passed to the underlying
-         :class:`.Table` object created for the command.
+         :class:`sqlalchemy.schema.Table` object created for the command.
 
         """
         self.impl.create_table(
             self._table(name, *columns, **kw)
         )
 
-    def drop_table(self, name):
+    def drop_table(self, name, **kw):
         """Issue a "drop table" instruction using the current
         migration context.
 
@@ -536,12 +640,20 @@ class Operations(object):
 
             drop_table("accounts")
 
+        :param name: Name of the table
+        :param schema: Optional schema name to operate within.
+
+         .. versionadded:: 0.4.0
+
+        :param \**kw: Other keyword arguments are passed to the underlying
+         :class:`sqlalchemy.schema.Table` object created for the command.
+
         """
         self.impl.drop_table(
-            self._table(name)
+            self._table(name, **kw)
         )
 
-    def create_index(self, name, tablename, *columns, **kw):
+    def create_index(self, name, table_name, columns, schema=None, **kw):
         """Issue a "create index" instruction using the current
         migration context.
 
@@ -550,52 +662,94 @@ class Operations(object):
             from alembic import op
             op.create_index('ik_test', 't1', ['foo', 'bar'])
 
+        :param name: name of the index.
+        :param table_name: name of the owning table.
+
+         .. versionchanged:: 0.5.0
+            The ``tablename`` parameter is now named ``table_name``.
+            As this is a positional argument, the old name is no
+            longer present.
+
+        :param columns: a list of string column names in the
+         table.
+        :param schema: Optional schema name to operate within.
+
+         .. versionadded:: 0.4.0
+
         """
 
         self.impl.create_index(
-            self._index(name, tablename, *columns, **kw)
+            self._index(name, table_name, columns, schema=schema, **kw)
         )
 
-    def drop_index(self, name, tablename=None):
+    @util._with_legacy_names([('tablename', 'table_name')])
+    def drop_index(self, name, table_name=None, schema=None):
         """Issue a "drop index" instruction using the current
         migration context.
 
-
         e.g.::
 
             drop_index("accounts")
 
-        :param tablename: name of the owning table.  Some
+        :param name: name of the index.
+        :param table_name: name of the owning table.  Some
          backends such as Microsoft SQL Server require this.
 
+         .. versionchanged:: 0.5.0
+            The ``tablename`` parameter is now named ``table_name``.
+            The old name will continue to function for backwards
+            compatibility.
+
+        :param schema: Optional schema name to operate within.
+
+         .. versionadded:: 0.4.0
+
         """
         # need a dummy column name here since SQLAlchemy
         # 0.7.6 and further raises on Index with no columns
-        self.impl.drop_index(self._index(name, tablename, ['x']))
+        self.impl.drop_index(
+            self._index(name, table_name, ['x'], schema=schema)
+        )
 
-    def drop_constraint(self, name, tablename, type=None):
+    @util._with_legacy_names([("type", "type_")])
+    def drop_constraint(self, name, table_name, type_=None, schema=None):
         """Drop a constraint of the given name, typically via DROP CONSTRAINT.
 
         :param name: name of the constraint.
-        :param tablename: tablename.
-        :param type: optional, required on MySQL.  can be
+        :param table_name: table name.
+
+         .. versionchanged:: 0.5.0
+            The ``tablename`` parameter is now named ``table_name``.
+            As this is a positional argument, the old name is no
+            longer present.
+
+        :param ``type_``: optional, required on MySQL.  can be
          'foreignkey', 'primary', 'unique', or 'check'.
 
-        .. versionadded:: 0.3.6 'primary' qualfier to enable
-           dropping of MySQL primary key constraints.
+         .. versionchanged:: 0.5.0
+            The ``type`` parameter is now named ``type_``.  The old name
+            ``type`` will remain for backwards compatibility.
+
+         .. versionadded:: 0.3.6 'primary' qualfier to enable
+            dropping of MySQL primary key constraints.
+
+        :param schema: Optional schema name to operate within.
+
+         .. versionadded:: 0.4.0
 
         """
-        t = self._table(tablename)
+
+        t = self._table(table_name, schema=schema)
         types = {
-            'foreignkey':lambda name:schema.ForeignKeyConstraint(
+            'foreignkey': lambda name: sa_schema.ForeignKeyConstraint(
                                 [], [], name=name),
-            'primary':schema.PrimaryKeyConstraint,
-            'unique':schema.UniqueConstraint,
-            'check':lambda name:schema.CheckConstraint("", name=name),
-            None:schema.Constraint
+            'primary': sa_schema.PrimaryKeyConstraint,
+            'unique': sa_schema.UniqueConstraint,
+            'check': lambda name: sa_schema.CheckConstraint("", name=name),
+            None: sa_schema.Constraint
         }
         try:
-            const = types[type]
+            const = types[type_]
         except KeyError:
             raise TypeError("'type' can be one of %s" %
                         ", ".join(sorted(repr(x) for x in types)))
@@ -665,7 +819,7 @@ class Operations(object):
          numerics should be supported.   Other types like boolean,
          dates, etc. may or may not be supported yet by various
          backends.
-        :param type_: optional - a :class:`sqlalchemy.types.TypeEngine`
+        :param ``type_``: optional - a :class:`sqlalchemy.types.TypeEngine`
          subclass stating the type of this value.  In SQLAlchemy
          expressions, this is usually derived automatically
          from the Python type of the value itself, as well as
@@ -734,7 +888,7 @@ class Operations(object):
 
         :param execution_options: Optional dictionary of
          execution options, will be passed to
-         :meth:`sqlalchemy.engine.base.Connection.execution_options`.
+         :meth:`sqlalchemy.engine.Connection.execution_options`.
         """
         self.migration_context.impl.execute(sql,
                     execution_options=execution_options)
@@ -743,7 +897,7 @@ class Operations(object):
         """Return the current 'bind'.
 
         Under normal circumstances, this is the
-        :class:`~sqlalchemy.engine.base.Connection` currently being used
+        :class:`~sqlalchemy.engine.Connection` currently being used
         to emit SQL to the database.
 
         In a SQL script context, this value is ``None``. [TODO: verify this]
