@@ -1,38 +1,35 @@
-from __future__ import with_statement
-
-from mako.template import Template
 import sys
 import os
 import textwrap
-from sqlalchemy.engine import url
-from sqlalchemy import util as sqla_util
-import imp
 import warnings
 import re
 import inspect
-import time
-import random
 import uuid
 
+from mako.template import Template
+from sqlalchemy.engine import url
+from sqlalchemy import __version__
+
+from .compat import callable, exec_, load_module
+
 class CommandError(Exception):
     pass
 
-from sqlalchemy import __version__
-_vers = tuple([int(x) for x in __version__.split(".")])
-sqla_06 = _vers > (0, 6)
-sqla_07 = _vers > (0, 7)
-if not sqla_06:
+def _safe_int(value):
+    try:
+        return int(value)
+    except:
+        return value
+_vers = tuple([_safe_int(x) for x in re.findall(r'(\d+|[abc]\d)', __version__)])
+sqla_07 = _vers > (0, 7, 2)
+sqla_08 = _vers >= (0, 8, 0, 'b2')
+if not sqla_07:
     raise CommandError(
-            "SQLAlchemy 0.6 or greater is required. "
-            "Version 0.7 or above required for full featureset.")
+            "SQLAlchemy 0.7.3 or greater is required. ")
+
+from sqlalchemy.util import format_argspec_plus, update_wrapper
+from sqlalchemy.util.compat import inspect_getfullargspec
 
-def requires_07(feature):
-    if not sqla_07:
-        raise CommandError(
-            "The %s feature requires "
-            "SQLAlchemy 0.7 or greater."
-            % feature
-        )
 try:
     width = int(os.environ['COLUMNS'])
 except (KeyError, ValueError):
@@ -47,7 +44,7 @@ def template_to_file(template_file, dest, **kw):
 def create_module_class_proxy(cls, globals_, locals_):
     """Create module level proxy functions for the
     methods on a given class.
-    
+
     The functions will have a compatible signature
     as the methods.   A proxy is established
     using the ``_install_proxy(obj)`` function,
@@ -81,12 +78,12 @@ def create_module_class_proxy(cls, globals_, locals_):
             num_defaults += len(spec[3])
         name_args = spec[0]
         if num_defaults:
-            defaulted_vals = name_args[0-num_defaults:]
+            defaulted_vals = name_args[0 - num_defaults:]
         else:
             defaulted_vals = ()
 
         apply_kw = inspect.formatargspec(
-                                name_args, spec[1], spec[2], 
+                                name_args, spec[1], spec[2],
                                 defaulted_vals,
                                 formatvalue=lambda x: '=' + x)
 
@@ -110,13 +107,13 @@ def create_module_class_proxy(cls, globals_, locals_):
             return _proxy.%(name)s(%(apply_kw)s)
             e
         """ % {
-            'name':name,
-            'args':args[1:-1],
-            'apply_kw':apply_kw[1:-1],
-            'doc':fn.__doc__,
+            'name': name,
+            'args': args[1:-1],
+            'apply_kw': apply_kw[1:-1],
+            'doc': fn.__doc__,
         })
         lcl = {}
-        exec func_text in globals_, lcl
+        exec_(func_text, globals_, lcl)
         return lcl[name]
 
     for methname in dir(cls):
@@ -128,10 +125,10 @@ def create_module_class_proxy(cls, globals_, locals_):
 
 def coerce_resource_to_filename(fname):
     """Interpret a filename as either a filesystem location or as a package resource.
-    
+
     Names that are non absolute paths and contain a colon
     are interpreted as resources and coerced to a file location.
-    
+
     """
     if not os.path.isabs(fname) and ":" in fname:
         import pkg_resources
@@ -158,6 +155,10 @@ def obfuscate_url_pw(u):
         u.password = 'XXXXX'
     return str(u)
 
+def asbool(value):
+    return value is not None and \
+        value.lower() == 'true'
+
 def warn(msg):
     warnings.warn(msg)
 
@@ -165,7 +166,7 @@ def msg(msg, newline=True):
     lines = textwrap.wrap(msg, width)
     if len(lines) > 1:
         for line in lines[0:-1]:
-            sys.stdout.write("  " +line + "\n")
+            sys.stdout.write("  " + line + "\n")
     sys.stdout.write("  " + lines[-1] + ("\n" if newline else ""))
 
 def load_python_file(dir_, filename):
@@ -173,17 +174,17 @@ def load_python_file(dir_, filename):
 
     module_id = re.sub(r'\W', "_", filename)
     path = os.path.join(dir_, filename)
-    module = imp.load_source(module_id, path, open(path, 'rb'))
+    module = load_module(module_id, path)
     del sys.modules[module_id]
     return module
 
 def pyc_file_from_path(path):
     """Given a python source path, locate the .pyc.
-    
+
     See http://www.python.org/dev/peps/pep-3147/
                         #detecting-pep-3147-availability
         http://www.python.org/dev/peps/pep-3147/#file-extension-checks
-    
+
     """
     import imp
     has3147 = hasattr(imp, 'get_tag')
@@ -241,3 +242,42 @@ class immutabledict(dict):
 
     def __repr__(self):
         return "immutabledict(%s)" % dict.__repr__(self)
+
+
+def _with_legacy_names(translations):
+    def decorate(fn):
+
+        spec = inspect_getfullargspec(fn)
+        metadata = dict(target='target', fn='fn')
+        metadata.update(format_argspec_plus(spec, grouped=False))
+
+        has_keywords = bool(spec[2])
+
+        if not has_keywords:
+            metadata['args'] += ", **kw"
+            metadata['apply_kw'] += ", **kw"
+
+        def go(*arg, **kw):
+            names = set(kw).difference(spec[0])
+            for oldname, newname in translations:
+                if oldname in kw:
+                    kw[newname] = kw.pop(oldname)
+                    names.discard(oldname)
+
+                    warnings.warn(
+                        "Argument '%s' is now named '%s' for function '%s'" %
+                        (oldname, newname, fn.__name__))
+            if not has_keywords and names:
+                raise TypeError("Unknown arguments: %s" % ", ".join(names))
+            return fn(*arg, **kw)
+
+        code = 'lambda %(args)s: %(target)s(%(apply_kw)s)' % (
+                metadata)
+        decorated = eval(code, {"target": go})
+        decorated.__defaults__ = getattr(fn, '__func__', fn).__defaults__
+        return update_wrapper(decorated, fn)
+
+    return decorate
+
+
+
